# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Wrapper functions for the _image_enhance C extension."""

from gi.repository import GdkPixbuf, GLib
from vimiv import _image_enhance


def enhance_bc(pixbuf, brightness, contrast):
    """Enhance brightness and contrast of a GdkPixbuf.Pixbuf.

    Args:
        pixbuf: Original GdkPixbuf.Pixbuf to work with.
        brightness: Float between -1.0 and 1.0 to change brightness.
        contrast: Float between -1.0 and 1.0 to change contrast.
    Return:
        The enhanced GdkPixbuf.Pixbuf
    """
    data = pixbuf.get_pixels()
    has_alpha = pixbuf.get_has_alpha()
    c_has_alpha = 1 if has_alpha else 0  # Numbers are easier for C
    # Update plain bytes using C extension
    # Pylint does not read this properly
    # pylint: disable=no-member
    data = _image_enhance.enhance_bc(data, c_has_alpha, brightness, contrast)
    gdata = GLib.Bytes.new(data)
    return GdkPixbuf.Pixbuf.new_from_bytes(gdata,
                                           pixbuf.get_colorspace(),
                                           has_alpha,
                                           pixbuf.get_bits_per_sample(),
                                           pixbuf.get_width(),
                                           pixbuf.get_height(),
                                           pixbuf.get_rowstride())



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""All parsers for vimiv."""

import configparser
import os
import sys

from gi.repository import GLib
from vimiv.exceptions import SettingNotFoundError, StringConversionError
from vimiv.helpers import error_message
from vimiv.settings import settings


def get_aliases(config):
    """Add aliases from the configfile to the ALIASES section of settings.

    Args:
        config: configparser.ConfigParser of the configfile.

    Return:
        aliases: Dictionary of aliases.
        message: Error message filled with aliases that cannot be parsed.
    """
    message = ""
    aliases = {}

    try:
        alias_section = config["ALIASES"]
    except KeyError:
        # Default to no aliases if the section does not exist in the configfile
        alias_section = dict()

    for alias in alias_section:
        try:
            aliases[alias] = alias_section[alias]
        except configparser.InterpolationError as e:
            message += "Parsing alias '%s' failed.\n" % alias \
                + "If you meant to use % for current file, use %%.\n" \
                + e.message + "\n"

    return aliases, message


def parse_config(commandline_config=None, running_tests=False):
    """Check each configfile for settings and apply them.

    Args:
        commandline_config: Configfile given by command line flag to parse.
        running_tests: If True, running from testsuite.
    Return:
        Dictionary of modified settings.
    """
    configfiles = []
    # We do not want to parse user configuration files when running the test
    # suite
    if not running_tests:
        configfiles += [
            "/etc/vimiv/vimivrc",
            os.path.join(GLib.get_user_config_dir(), "vimiv/vimivrc"),
            os.path.expanduser("~/.vimiv/vimivrc")]
    if commandline_config:
        configfiles.append(commandline_config)

    # Error message, gets filled with invalid sections in the user's configfile.
    # If any exist, a popup is displayed at startup.
    message = ""

    # Let ConfigParser parse the list of configuration files
    config = configparser.ConfigParser()
    try:
        config.read(configfiles)
    except UnicodeDecodeError as e:
        message += "Could not decode configfile.\n" + str(e)
    except configparser.MissingSectionHeaderError as e:
        message += "Invalid configfile.\n" + str(e)
    except configparser.ParsingError as e:
        message += str(e)

    # Override settings with settings in config
    for header in ["GENERAL", "LIBRARY", "EDIT"]:
        if header not in config:
            continue
        section = config[header]
        for setting in section:
            try:
                settings.override(setting, section[setting])
            except StringConversionError as e:
                message += str(e) + "\n"
            except SettingNotFoundError:
                message += "Unknown setting %s\n" % (setting)

    # Receive aliases
    aliases, partial_message = get_aliases(config)
    settings.set_aliases(aliases)
    message += partial_message

    if message:
        error_message(message, running_tests=running_tests)
    return settings


def parse_keys(keyfiles=None, running_tests=False):
    """Check for a keyfile and parse it.

    Args:
        keyfiles: List of keybinding files. If not None, use this list instead
            of the default files.
        running_tests: If True running from testsuite. Do not show error popup.
    Return:
        Dictionary of keybindings.
    """
    if not keyfiles:
        keyfiles = \
            ["/etc/vimiv/keys.conf",
             os.path.join(GLib.get_user_config_dir(), "vimiv/keys.conf"),
             os.path.expanduser("~/.vimiv/keys.conf")]
    # Read the list of files
    keys = configparser.ConfigParser()
    try:
        # No file for keybindings found
        if not keys.read(keyfiles):
            message = "Keyfile not found. Exiting."
            error_message(message, running_tests=running_tests)
            sys.exit(1)
    except configparser.DuplicateOptionError as e:
        message = e.message + ".\n Duplicate keybinding. Exiting."
        error_message(message, running_tests=running_tests)
        sys.exit(1)

    # Get the keybinding dictionaries checking for errors
    try:
        keys_image = keys["IMAGE"]
        keys_thumbnail = keys["THUMBNAIL"]
        keys_library = keys["LIBRARY"]
        keys_manipulate = keys["MANIPULATE"]
        keys_command = keys["COMMAND"]
    except KeyError as e:
        message = "Missing section " + str(e) + " in keys.conf.\n" \
                  "Refer to vimivrc(5) to fix your config."
        error_message(message, running_tests=running_tests)
        sys.exit(1)

    # Update the dictionaries of every window with the keybindings that apply
    # for more than one window
    def update_keybindings(sections, keydict):
        """Add keybindings from generic sections to keydict."""
        for section in sections:
            if section in keys:
                print("Section", section, "is deprecated and will be removed in"
                      " a future version.")
                keydict.update(keys[section])
    update_keybindings(["GENERAL", "IM_THUMB", "IM_LIB"], keys_image)
    update_keybindings(["GENERAL", "IM_THUMB"], keys_thumbnail)
    update_keybindings(["GENERAL", "IM_LIB"], keys_library)

    # Generate one dictionary for all and return it
    keybindings = {"IMAGE": keys_image,
                   "THUMBNAIL": keys_thumbnail,
                   "LIBRARY": keys_library,
                   "MANIPULATE": keys_manipulate,
                   "COMMAND": keys_command}
    return keybindings



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Contains all commands and functions for vimiv."""

from vimiv.exceptions import (AliasError, ArgumentAmountError,
                              StringConversionError, NotABoolean, NotANumber,
                              SettingNotFoundError)
from vimiv.fileactions import format_files
from vimiv.helpers import error_message
from vimiv.settings import settings
from vimiv.tags import TagHandler


class Commands(object):
    """Initialize and store the commands for vimiv.

    Commands are added via Commands.add_command() and can be received via
    Commands["command_name"].

    Attributes:
        aliases: Dictionary of stored aliases.

        _app: The main vimiv application to interact with.
        _commands: Dictionary of stored commands.
        _n: Integer for iterator.
    """

    # Adding one command per statement is absolutely acceptable
    # pylint:disable=too-many-statements
    def __init__(self, app):
        """Generate commands and functions.

        Args:
            app: The main vimiv application to interact with.
        """
        self._app = app
        self._commands = {}
        self.aliases = {}
        self._n = 0
        self._app["tag_handler"] = TagHandler(app)
        # Add all commands
        self.add_command("accept_changes",
                         self._app["manipulate"].finish,
                         default_args=[True])
        self.add_command("alias", self.add_alias,
                         positional_args=["name", "command"],
                         last_arg_allows_space=True)
        self.add_command("autorotate", self._app["transform"].rotate_auto)
        self.add_command("center", self._app["main_window"].center_window)
        self.add_command("copy_basename", self._app["clipboard"].copy_name,
                         default_args=[False])
        self.add_command("copy_abspath", self._app["clipboard"].copy_name,
                         default_args=[True])
        self.add_command("delete", self._app["transform"].delete)
        self.add_command("undelete", self._app["transform"].undelete,
                         positional_args=["basename"],
                         last_arg_allows_space=True)
        self.add_command("discard_changes",
                         self._app["manipulate"].finish,
                         default_args=[False])
        self.add_command("edit", self._app["manipulate"].cmd_edit,
                         positional_args=["manipulation"],
                         optional_args=["value"])
        self.add_command("first", self._app["image"].move_pos,
                         default_args=[False], supports_count=True)
        self.add_command("first_lib", self._app["library"].move_pos,
                         default_args=[False], supports_count=True)
        self.add_command("fit", self._app["image"].zoom_to,
                         default_args=[0, "fit"])
        self.add_command("fit_horiz", self._app["image"].zoom_to,
                         default_args=[0, "horizontal"])
        self.add_command("fit_vert", self._app["image"].zoom_to,
                         default_args=[0, "vertical"])
        self.add_command("flip", self._app["transform"].flip,
                         positional_args=["direction"])
        self.add_command("format", format_files, default_args=[self._app],
                         positional_args=["formatstring"],
                         last_arg_allows_space=True)
        self.add_command("fullscreen", self._app["window"].toggle_fullscreen)
        self.add_command("last", self._app["image"].move_pos,
                         default_args=[True], supports_count=True)
        self.add_command("last_lib", self._app["library"].move_pos,
                         default_args=[True], supports_count=True)
        self.add_command("library", self._app["library"].toggle)
        self.add_command("focus_library", self._app["library"].focus,
                         default_args=[True])
        self.add_command("unfocus_library", self._app["library"].focus,
                         default_args=[False])
        self.add_command("manipulate", self._app["manipulate"].toggle)
        self.add_command("mark", self._app["mark"].mark)
        self.add_command("mark_all", self._app["mark"].mark_all)
        self.add_command("mark_between", self._app["mark"].mark_between)
        self.add_command("mark_toggle", self._app["mark"].toggle_mark)
        self.add_command("move_up", self._app["library"].move_up,
                         supports_count=True)
        self.add_command("next", self._app["image"].move_index,
                         default_args=[True, True], supports_count=True)
        self.add_command("next!", self._app["image"].move_index,
                         default_args=[True, True, 1, True],
                         supports_count=True)
        self.add_command("prev", self._app["image"].move_index,
                         default_args=[False, True], supports_count=True)
        self.add_command("prev!", self._app["image"].move_index,
                         default_args=[False, True, 1, True],
                         supports_count=True)
        self.add_command("q", self._app.quit_wrapper)
        self.add_command("q!", self._app.quit_wrapper,
                         default_args=[True])
        self.add_command("reload_lib", self._app["library"].reload,
                         default_args=["."])
        self.add_command("rotate", self._app["transform"].rotate,
                         positional_args=["value"], supports_count=True)
        self.add_command("set", self.set, positional_args=["setting"],
                         optional_args=["value"], supports_count=True)
        self.add_command("slideshow", self._app["slideshow"].toggle,
                         supports_count=True)
        self.add_command("tag_load", self._app["tags"].load,
                         positional_args=["tagname"],
                         last_arg_allows_space=True)
        self.add_command("tag_remove", self._app["tags"].remove,
                         positional_args=["tagname"],
                         last_arg_allows_space=True)
        self.add_command("tag_write", self._app["tags"].write,
                         default_args=[self._app["mark"].marked],
                         positional_args=["tagname"],
                         last_arg_allows_space=True)
        self.add_command("thumbnail", self._app["thumbnail"].toggle)
        self.add_command("version", self._app["information"].show_version_info)
        self.add_command("w", self._app["transform"].write)
        self.add_command("wq", self._app["transform"].write,
                         default_args=[True])
        self.add_command("zoom_in", self._app["window"].zoom,
                         default_args=[True], optional_args=["steps"],
                         supports_count=True)
        self.add_command("zoom_out", self._app["window"].zoom,
                         default_args=[False], optional_args=["steps"],
                         supports_count=True)
        self.add_command("zoom_to", self._app["image"].zoom_to,
                         optional_args=["percent"], supports_count=True)
        # Hidden commands
        self.add_command("clear_status", self._app["statusbar"].clear_status,
                         is_hidden=True)
        self.add_command("command", self._app["commandline"].enter,
                         optional_args=["text"], is_hidden=True)
        self.add_command("complete", self._app["completions"].complete,
                         default_args=[False], is_hidden=True)
        self.add_command("complete_inverse", self._app["completions"].complete,
                         default_args=[True], is_hidden=True)
        self.add_command("discard_command", self._app["commandline"].leave,
                         default_args=[True], is_hidden=True)
        self.add_command("focus_slider", self._app["manipulate"].focus_slider,
                         positional_args=["name"], is_hidden=True)
        self.add_command("history_down",
                         self._app["commandline"].history_search,
                         default_args=[True], is_hidden=True)
        self.add_command("history_up", self._app["commandline"].history_search,
                         default_args=[False], is_hidden=True)
        self.add_command("scroll", self._app["main_window"].scroll,
                         positional_args=["direction"], supports_count=True,
                         is_hidden=True)
        self.add_command("scroll_lib", self._app["library"].scroll,
                         positional_args=["direction"], supports_count=True,
                         is_hidden=True)
        self.add_command("search", self._app["commandline"].enter_search,
                         is_hidden=True)
        self.add_command("search_next",
                         self._app["commandline"].search_move,
                         default_args=[True], supports_count=True,
                         is_hidden=True)
        self.add_command("search_prev",
                         self._app["commandline"].search_move,
                         default_args=[False], supports_count=True,
                         is_hidden=True)
        self.add_command("slider", self._app["manipulate"].change_slider,
                         optional_args=["value"], supports_count=True,
                         is_hidden=True)
        # Aliases
        aliases = settings.get_aliases()
        message = ""
        for alias in aliases:
            try:
                self._add_alias(alias, aliases[alias])
            except AliasError as e:
                message += str(e) + "\n"
        if message:
            error_message(message, running_tests=self._app.running_tests)
        aliases.clear()

    def set(self, setting, value=None):
        """Wrapper for settings.override catching errors."""
        # Allow passing multipliers via keybinding
        multiplier = self._app["eventhandler"].get_num_str()
        if multiplier:
            self._app["eventhandler"].num_clear()
        else:
            multiplier = "1"
        try:
            # Toggle boolean settings
            if setting.endswith("!"):
                setting = setting.rstrip("!")
                settings.toggle(setting)
            # Add to number settings
            elif value and value.startswith("+") or value.startswith("-"):
                settings.add_to(setting, value, multiplier)
            else:
                settings.override(setting, value)
        except (StringConversionError, SettingNotFoundError, NotABoolean,
                NotANumber) as e:
            self._app["statusbar"].message(str(e), "error")

    def __getitem__(self, name):
        """Return the command corresponding to name.

        Args:
            name: The name of the command to access.
        Return:
            Dictionary of the command information.
        """
        return self._commands[name]

    def __iter__(self):
        self._n = 0
        return self

    def __next__(self):
        """Iterate over commands."""
        items = list(self._commands.values())
        if (self._n - 1) < len(items) - 1:
            self._n += 1
            return items[self._n - 1]
        else:
            raise StopIteration

    def add_command(self, name, function, default_args=None,
                    positional_args=None, optional_args=None,
                    supports_count=False, is_hidden=False,
                    last_arg_allows_space=False):
        """Add a command.

        Args:
            name: Name of the command.
            function: Function that the command should call.
            default_args: Default arguments to pass to the function.
            positional_args: Arguments that must be passed by the user.
            optional_args: Arguments that may be passed by the user.
            supports_count: If false, delete any num_str before running the
                command.
            is_hidden: If true, hide command from the command line.
            last_arg_allows_space: If True, last argument is a type string which
                may contain whitespace.
        """
        command = Command(name, function)
        command.set_default_args(default_args)
        command.set_positional_args(positional_args)
        command.set_optional_args(optional_args)
        command.set_supports_count(supports_count)
        command.set_is_hidden(is_hidden)
        command.set_last_arg_allows_space(last_arg_allows_space)
        self._commands[name] = command

    def add_alias(self, name, command_name):
        """Wrapper for self._add_alias which displays error messages.

        Args:
            name: Name of the alias.
            command_name: Name of the original command.
        """
        try:
            self._add_alias(name, command_name)
        except AliasError as e:
            self._app["statusbar"].message(str(e), "error")

    def _add_alias(self, name, command_name):
        """Add an alias.

        Args:
            name: Name of the alias.
            command_name: Name of the original command.
        """
        command = command_name.split()[0]
        if name in self._commands.keys():
            raise AliasError('Alias "%s" would overwrite an existing command'
                             % (name))
        if command not in self._commands.keys() \
                and command_name[0] not in "!~/.":
            raise AliasError('Alias "%s" failed: no command called "%s"'
                             % (name, command_name))
        self.aliases[name] = command_name


class Command(object):
    """Stores a single command and its attributes.

    Attributes:
        name: Name of the command as used in the command line.
        supports_count: If false, delete any num_str before running the
            command.
        is_hidden: If true, hide command from the command line.

        _function: Function the command should call.
        _default_args: Default arguments to pass to the function.
        _positional_args: Arguments that must be passed by the user.
        _optional_args: Arguments that may be passed by the user.
        _last_arg_allows_space: If True, last argument is a type string which
            may contain whitespace.
    """

    def __init__(self, name, function):
        """Generate command and set all attributes.

        Args:
            name: Name of the command as used in the command line.
            function: Function the command should call.
        """
        self.name = name
        self.supports_count = False
        self.is_hidden = False
        self._function = function
        self._default_args = None
        self._positional_args = None
        self._optional_args = None
        self._last_arg_allows_space = False

    def set_default_args(self, args):
        self._default_args = args

    def set_positional_args(self, args):
        self._positional_args = args

    def set_optional_args(self, args):
        self._optional_args = args

    def set_supports_count(self, supports_count):
        self.supports_count = supports_count

    def set_is_hidden(self, is_hidden):
        self.is_hidden = is_hidden

    def set_last_arg_allows_space(self, last_arg_allows_space):
        self._last_arg_allows_space = last_arg_allows_space

    def _get_max_args(self):
        """Return the maximum amount of arguments a command may receive.

        Return:
            The maximum of arguments as integer.
        """
        num = 0
        if self._positional_args:
            num += len(self._positional_args)
        if self._optional_args:
            num += len(self._optional_args)
        return num

    def _get_min_args(self):
        return len(self._positional_args) if self._positional_args else 0

    def run(self, arguments):
        """Run the command.

        This checks if the given arguments are compatible with the command,
        converts them properly and then calls self.function.

        Args:
            arguments: Arguments passed to the command.
        """
        # Too less arguments
        if len(arguments) < self._get_min_args():
            message = "Missing positional arguments for command %s: %s" \
                % (self.name, ", ".join(self._positional_args))
            raise ArgumentAmountError(message)
        # Join last arguments to one string?
        if self._last_arg_allows_space:
            last_args = arguments[self._get_min_args() - 1:]
            last_arg = " ".join(last_args)
            arguments = arguments[:self._get_min_args() - 1] + [last_arg]
        # Too many arguments
        if len(arguments) > self._get_max_args():
            message = "Too many arguments for command " + self.name
            raise ArgumentAmountError(message)
        # Run
        if self._default_args:
            arguments = self._default_args + arguments
        self._function(*arguments)



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Provides class to handle a shared trash directory.

The TrashManager deletes and undeletes images from the user's Trash directory in
$XDG_DATA_HOME/Trash according to the freedesktop.org trash specification.
"""

import configparser
import os
import shutil
import tempfile
import time

from gi.repository import GLib
from vimiv.exceptions import TrashUndeleteError


class TrashManager():
    """Provides mechanism to delete and undelete images.

    Attributes:
        files_directory: Directory to which the "deleted" files are moved.
        info_directory: Directory in which information on the "deleted" files is
            stored.
    """

    def __init__(self):
        """Create a new TrashManager."""
        self.files_directory = os.path.join(GLib.get_user_data_dir(),
                                            "Trash/files")
        self.info_directory = os.path.join(GLib.get_user_data_dir(),
                                           "Trash/info")
        os.makedirs(self.files_directory, exist_ok=True)
        os.makedirs(self.info_directory, exist_ok=True)

    def delete(self, filename):
        """Move a file to the trash directory.

        Args:
            filename: The original name of the file.
        """
        trash_filename = self._get_trash_filename(filename)
        self._create_info_file(trash_filename, filename)
        shutil.move(filename, trash_filename)

    def undelete(self, basename):
        """Undelete a file from the trash directory.

        Args:
            basename: The basename of the file in the trash directory.
        """
        info_filename = os.path.join(self.info_directory,
                                     basename + ".trashinfo")
        trash_filename = os.path.join(self.files_directory, basename)
        if not os.path.exists(info_filename) \
                or not os.path.exists(trash_filename):
            raise TrashUndeleteError("file does not exist")
        if os.path.isdir(trash_filename):
            raise TrashUndeleteError("directories are not supported")
        info = configparser.ConfigParser()
        info.read(info_filename)
        content = info["Trash Info"]
        original_filename = content["Path"]
        # Directory of the file is not accessible
        if not os.path.isdir(os.path.dirname(original_filename)):
            raise TrashUndeleteError("original directory is not accessible")
        shutil.move(trash_filename, original_filename)
        os.remove(info_filename)

    def _get_trash_filename(self, filename):
        """Return the name of the file in self.files_directory.

        Args:
            filename: The original name of the file.
        """
        path = os.path.join(self.files_directory, os.path.basename(filename))
        # Ensure that we do not overwrite any files
        extension = 2
        original_path = path
        while os.path.exists(path):
            path = original_path + "." + str(extension)
            extension += 1
        return path

    def _create_info_file(self, trash_filename, original_filename):
        """Create file with information as specified by the standard.

        Args:
            trash_filename: The name of the file in self.files_directory.
            original_filename: The original name of the file.
        """
        # Note: we cannot use configparser here as it writes keys in lowercase
        info_path = os.path.join(
            self.info_directory,
            os.path.basename(trash_filename) + ".trashinfo")
        # Write to temporary file and use shutil.move to make sure the operation
        # is an atomic operation as specified by the standard
        fd, temp_path = tempfile.mkstemp(dir=self.info_directory)
        os.close(fd)
        temp_file = open(temp_path, "w")
        temp_file.write("[Trash Info]\n")
        temp_file.write("Path=%s\n" % (original_filename))
        temp_file.write("DeletionDate=%s\n" % (time.strftime("%Y%m%dT%H%M%S")))
        # Make sure that all data is on disk
        temp_file.flush()
        os.fsync(temp_file.fileno())
        temp_file.close()
        shutil.move(temp_path, info_path)

    def get_files_directory(self):
        return self.files_directory

    def get_info_directory(self):
        return self.info_directory



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Image part of vimiv."""

from random import shuffle
from threading import Thread

from gi.repository import GdkPixbuf, GLib, Gtk
from vimiv.exceptions import StringConversionError
from vimiv.fileactions import is_animation, is_svg
from vimiv.helpers import get_float
from vimiv.settings import settings


class Image(Gtk.Image):
    """Image class for vimiv.

    Inherits from Gtk.Image and includes all actions that apply to it.

    Attributes:
        fit_image:
            user: Image is user zoomed.
            overzoom: Image is zoomed to fit respecting overzoom.
            horizontal: Image is zoomed to fit horizontally.
            vertical: Image is zoomed to fit vertically.
            fit: Image is zoomed to fit.
        zoom_percent: Percentage to zoom to compared to the original size.

        _app: The main vimiv class to interact with.
        _identifier: Used so GUI callbacks are only done if the image is equal
        _pixbuf_iter: Iter of displayed animation.
        _pixbuf_original: Original image.
        _size: Size of the displayed image as a tuple.
        _timer_id: Id of current animation timer.
        _faulty_image: Necessary evil for images that PixbufLoader cannot read.
    """

    def __init__(self, app):
        """Set default values for attributes."""
        super(Image, self).__init__()
        self._app = app

        # Settings and defaults
        self.fit_image = "overzoom"
        self._pixbuf_iter = GdkPixbuf.PixbufAnimationIter()
        self._pixbuf_original = GdkPixbuf.Pixbuf()
        self.zoom_percent = 1
        self._identifier = 0
        self._size = (1, 1)
        self._timer_id = 0
        self._faulty_image = False

        # Connect signals
        self._app["transform"].connect("changed", self._on_image_changed)
        self._app["commandline"].search.connect("search-completed",
                                                self._on_search_completed)
        settings.connect("changed", self._on_settings_changed)

    def _update(self):
        """Show the final image."""
        if not self._app.get_paths() or self._faulty_image:
            return
        # Scale image
        pbo_width = self._pixbuf_original.get_width()
        pbo_height = self._pixbuf_original.get_height()
        pbf_width = int(pbo_width * self.zoom_percent)
        pbf_height = int(pbo_height * self.zoom_percent)
        # Rescaling of svg
        if is_svg(self._app.get_path()) and settings["rescale_svg"].get_value():
            pixbuf_final = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                self._app.get_path(), -1, pbf_height, True)
        else:
            pixbuf_final = self._pixbuf_original.scale_simple(
                pbf_width, pbf_height, GdkPixbuf.InterpType.BILINEAR)
        self.set_from_pixbuf(pixbuf_final)
        # Update the statusbar
        self._app["statusbar"].update_info()

    def zoom_delta(self, zoom_in=True, step=1):
        """Zoom the image by delta percent.

        Args:
            zoom_in: If True zoom in, else zoom out.
        """
        delta = 0.25
        # Allow user steps
        step = self._app["eventhandler"].num_receive(step, True)
        if isinstance(step, str):
            try:
                step = get_float(step, allow_sign=True)
            except StringConversionError as e:
                self._app["statusbar"].message(str(e), "error")
                return
        fallback_zoom = self.zoom_percent
        if zoom_in:
            self.zoom_percent = self.zoom_percent * (1 + delta * step)
        else:
            self.zoom_percent = self.zoom_percent / (1 + delta * step)
        self._catch_unreasonable_zoom_and_update(fallback_zoom)
        self.fit_image = "user"

    def zoom_to(self, percent=0, fit="fit"):
        """Zoom to a given percentage.

        Args:
            percent: Percentage to zoom to.
            fit: How to fit image if percent is not given.
        """
        fallback_zoom = self.zoom_percent
        # Catch user zooms
        percent = self._app["eventhandler"].num_receive(percent, True)
        # Given from commandline
        if isinstance(percent, str):
            try:
                percent = get_float(percent)
            except StringConversionError as e:
                self._app["statusbar"].message(str(e), "error")
                return
        self._size = self._get_available_size()
        # 0 means zoom to fit
        if percent:
            self.zoom_percent = percent
            self.fit_image = "user"
        else:
            self.zoom_percent = self.get_zoom_percent_to_fit(fit)
            self.fit_image = fit
        # Catch some unreasonable zooms
        self._catch_unreasonable_zoom_and_update(fallback_zoom)

    def move_index(self, forward=True, key=None, delta=1, force=False):
        """Move by delta in paths.

        Args:
            forward: If True move forwards. Else move backwards.
            key: If available, move by key.
            delta: Positions to move by.
            force: If True, move regardless of editing image.
        """
        # Check if an image is opened or if it has been edited
        if not self._app.get_paths() or self._app["thumbnail"].toggled or \
                self._app["manipulate"].check_for_edit(force):
            return
        # Apply done rotations and flips to file
        self._app["transform"].apply()
        # Check for prepended numbers and direction
        if key:
            delta *= self._app["eventhandler"].num_receive()
        if not forward:
            delta *= -1
        self._app.update_index(delta)
        self.fit_image = "overzoom"

        # Reshuffle on wrap-around
        if settings["shuffle"].get_value() \
                and self._app.get_index() is 0 and delta > 0:
            shuffle(self._app.get_paths())

        # Load the image at path into self._pixbuf_* and show it
        self.load()

    def load(self):
        """Load an image using GdkPixbufLoader."""
        path = self._app.get_path()
        # Remove old timers and reset scale
        if self._timer_id:
            self.zoom_percent = 1
            self._pause_gif()
        # Load file
        try:
            self._load(path)
        except (PermissionError, FileNotFoundError):
            self._app.remove_path(path)
            self.move_pos(False)
            self._app["statusbar"].message("File not accessible", "error")

    def move_pos(self, forward=True, force=False):
        """Move to specific position in paths.

        Args:
            forward: If True move forwards. Else move backwards.
            force: If True, move regardless of editing image.
        """
        # Check if image has been edited
        if self._app["manipulate"].check_for_edit(force):
            return
        max_pos = len(self._app.get_paths())
        current = self._app.get_index()
        # Move to definition by keys or end/beg
        if forward:
            pos = self._app["eventhandler"].num_receive(max_pos)
        else:
            pos = self._app["eventhandler"].num_receive()
        # Catch range
        if pos < 0 or pos > max_pos:
            self._app["statusbar"].message("Unsupported index", "warning")
            return
        # Do the maths and move
        if self._app["thumbnail"].toggled:
            self._app["thumbnail"].move_to_pos(pos - 1)
        else:
            self.move_index(True, False, pos - current - 1)

    def get_scroll_scale(self):
        return self.zoom_percent / self.get_zoom_percent_to_fit() * 2

    def get_zoom_percent(self):
        return self.zoom_percent * 100

    def get_zoom_percent_to_fit(self, fit="fit"):
        """Get the zoom factor perfectly fitting the image to the window.

        Args:
            fit: How to fit image.
        Return:
            Zoom percentage.
        """
        # Size of the file
        pbo_width = self._pixbuf_original.get_width()
        pbo_height = self._pixbuf_original.get_height()
        # Maximum size respecting overzoom
        max_width = pbo_width * settings["overzoom"].get_value()
        max_height = pbo_height * settings["overzoom"].get_value()
        # Get scales for "panorama" vs "portrait" image
        w_scale = pbo_width / self._size[0]
        h_scale = pbo_height / self._size[1]
        scale_width = True if w_scale > h_scale else False
        # Check if image fits completely with overzoom
        fits = max_width < self._size[0] and max_height < self._size[1]
        # Image fits completely even with overzoom and we do not want to fit
        if fits and fit in ["user", "overzoom"]:
            return settings["overzoom"].get_value()
        # Force horizontal fit or "panorama" image
        elif fit == "horizontal" or (scale_width and fit != "vertical"):
            return self._size[0] / pbo_width
        # Force vertical fit or "portrait" image
        return self._size[1] / pbo_height

    def _catch_unreasonable_zoom_and_update(self, fallback_zoom):
        """Catch unreasonable zooms otherwise update.

        Args:
            fallback_zoom: Zoom percentage to fall back to if the zoom
                percentage is unreasonable.
        """
        orig_width = self._pixbuf_original.get_width()
        orig_height = self._pixbuf_original.get_height()
        new_width = orig_width * self.zoom_percent
        new_height = orig_height * self.zoom_percent
        min_width = max(16, orig_width * 0.05)
        min_height = max(16, orig_height * 0.05)
        max_width = min(self._app["window"].get_size()[0] * 10,
                        self._pixbuf_original.get_width() * 20)
        max_height = min(self._app["window"].get_size()[1] * 10,
                         self._pixbuf_original.get_height() * 20)
        # Image too small or too large
        if new_height < min_height or new_width < min_width \
                or new_height > max_height or new_width > max_width:
            # Warn user if it was his fault
            if self.fit_image == "user":
                message = "Image cannot be zoomed this far"
                self._app["statusbar"].message(message, "warning")
                self.zoom_percent = fallback_zoom
            return
        self._update()

    def _play_gif(self):
        """Run the animation of a gif."""
        self._pixbuf_original = self._pixbuf_iter.get_pixbuf()
        GLib.idle_add(self._update)
        if self._pixbuf_iter.advance():
            # Clear old timer
            if self._timer_id:
                GLib.source_remove(self._timer_id)
            # Add new timer if the gif is not static
            delay = self._pixbuf_iter.get_delay_time()
            self._timer_id = GLib.timeout_add(delay, self._play_gif) \
                if delay >= 0 else 0

    def _pause_gif(self):
        """Pause a gif or show initial image."""
        if self._timer_id:
            GLib.source_remove(self._timer_id)
            self._timer_id = 0
        else:
            image = self._pixbuf_iter.get_pixbuf()
            self.set_from_pixbuf(image)

    def _get_available_size(self):
        """Receive size not occupied by other Widgets.

        Subtracts other widgets (manipulate, statusbar, library) from window
        size and returns the available size.

        Return:
            Tuple of available size.
        """
        size = self._app["window"].get_size()
        if self._app["library"].grid.is_visible():
            library_width = self._app["library"].get_size_request()[0]
            size = (size[0] - library_width, size[1])
        if settings["display_bar"].get_value():
            size = (size[0], size[1] - self._app["statusbar"].get_bar_height())
        return size

    def _load(self, path):
        """Actual implementation to load an image from path."""
        loader = GdkPixbuf.PixbufLoader()
        self._identifier += 1
        if is_animation(path):
            loader.connect("area-prepared", self._set_image_anim)
        else:
            loader.connect("area-prepared", self._set_image_pixbuf)
            loader.connect("closed",
                           self._finish_image_pixbuf, self._identifier)
        load_thread = Thread(target=self._load_thread, args=(loader, path),
                             daemon=True)
        # Daemon is set to True so the program can exit with "q" immediately if
        # only a loading thread is left
        load_thread.start()

    def _load_thread(self, loader, path):
        # The try ... except wrapper and the _faulty_image attribute are used to
        # catch weird images that break GdkPixbufLoader but work otherwise
        # See https://github.com/karlch/vimiv/issues/49 for more information
        try:
            self._faulty_image = True
            with open(path, "rb") as f:
                image_bytes = f.read()
                loader.write(image_bytes)
            self._faulty_image = False
            loader.close()
        except GLib.GError:
            self._pixbuf_original = GdkPixbuf.Pixbuf.new_from_file(path)
            self._faulty_image = False
            self._set_image_pixbuf()
            GLib.idle_add(self._update)

    def _set_image_pixbuf(self, loader=None):
        if loader:
            self._pixbuf_original = loader.get_pixbuf()
        self._size = self._get_available_size()
        self.zoom_percent = self.get_zoom_percent_to_fit(self.fit_image)

    def _finish_image_pixbuf(self, loader, image_id):
        if self._identifier == image_id:
            GLib.idle_add(self._update)

    def _set_image_anim(self, loader):
        self._pixbuf_iter = loader.get_animation().get_iter()
        self._pixbuf_original = self._pixbuf_iter.get_pixbuf()
        self._size = self._get_available_size()
        self.zoom_percent = self.get_zoom_percent_to_fit(self.fit_image)
        if settings["play_animations"].get_value():
            delay = self._pixbuf_iter.get_delay_time()
            self._timer_id = GLib.timeout_add(delay, self._play_gif) \
                if delay >= 0 else 0

    def get_pixbuf_original(self):
        return self._pixbuf_original.copy()

    def set_pixbuf(self, pixbuf):
        self._pixbuf_original = pixbuf
        self._update()

    def _on_image_changed(self, transform, change, arg):
        """Update image after a transformation.

        Args:
            transform: The transform object that emitted the signal.
            change: The type of transformation.
            arg: Argument for the transformation, e.g. cwise for rotate.
        """
        if change == "rotate":
            self._pixbuf_original = \
                self._pixbuf_original.rotate_simple(90 * arg)
        elif change == "flip":
            self._pixbuf_original = self._pixbuf_original.flip(arg)
        if self.fit_image != "user":
            self.zoom_to(0, self.fit_image)
        else:
            self._update()

    def _on_search_completed(self, search, new_pos, last_focused):
        if last_focused == "im":
            self._app["eventhandler"].set_num_str(new_pos + 1)
            self.move_pos()

    def _on_settings_changed(self, new_settings, setting):
        # Rescale image after setting overzoom
        if setting == "overzoom" and self.fit_image != "user":
            self.zoom_percent = self.get_zoom_percent_to_fit(self.fit_image)
            self._update()
        # Change play status of gifs
        elif setting == "play_animations":
            if self._app.get_paths() and is_animation(self._app.get_path()) \
                    and not self._app["thumbnail"].toggled:
                if settings["play_animations"].get_value():
                    self._play_gif()
                else:
                    self._pause_gif()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Contains the commandline class for vimiv."""

import os
import re
from subprocess import PIPE, Popen
from threading import Thread

from gi.repository import GLib, GObject, Gtk
from vimiv.commands import Commands
from vimiv.exceptions import ArgumentAmountError, NoSearchResultsError
from vimiv.helpers import error_message, expand_filenames, read_file
from vimiv.settings import settings


class CommandLine(Gtk.Entry):
    """Commandline of vimiv.

    Attributes:
        commands: Commands class storing all commands.
        search: Search class defined below.
        running_processes: List of all running external processes.

        _app: The main vimiv application to interact with.
        _history: List of commandline history to save.
        _last_widget: Widget that was focused before the command line.
        _last_index: Index that was last selected in the library, if any.
        _matching_history: Parts of the history that match entered text.
        _pos: Index in _history list.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        super(CommandLine, self).__init__()
        self._app = app
        self.commands = None

        # Command line
        self.connect("activate", self._on_activate)
        self.connect("key_press_event",
                     self._app["eventhandler"].on_key_press, "COMMAND")
        self.connect("changed", self._on_text_changed)
        self.set_hexpand(True)

        # Initialize search
        self.search = Search(self)

        # Cmd history from file
        datadir = os.path.join(GLib.get_user_data_dir(), "vimiv")
        os.makedirs(datadir, exist_ok=True)
        # If there is a history file in the old location move it and inform the
        # user
        # TODO remove this in a new version, assigned for 0.10
        old_histfile = os.path.expanduser("~/.vimiv/history")
        new_histfile = os.path.join(datadir, "history")
        if os.path.isfile(old_histfile):
            if os.path.isfile(new_histfile):
                message = "There is a history file in the deprecated location" \
                    " %s and the new location %s. Using the new one. If you " \
                    "want the history from the old file you will have to " \
                    "merge it manually." % (old_histfile, new_histfile)
                error_message(message, running_tests=self._app.running_tests)
            else:
                os.rename(old_histfile, new_histfile)
                message = "Moved the old history file %s " \
                    "to the new location %s." % (old_histfile, new_histfile)
                error_message(message, running_tests=self._app.running_tests)
        self._history = read_file(os.path.join(datadir, "history"))

        # Defaults
        self._matching_history = []
        self._last_index = 0
        self._pos = 0
        self.running_processes = []
        self._last_widget = ""

    def _on_activate(self, entry):
        """Handle input from the entry when activated.

        Check for type of command (internal, external or path) and run the
        correct function.

        Args:
            entry: The Gtk.Entry from which the command comes.
        """
        # Only close completions if currently tabbing through results
        if self._app["completions"].activate_tab_completion():
            return
        # cmd from input
        command = entry.get_text()
        # And close the cmd line by emptying text
        self.set_text("")
        # Write command to log in debug mode
        if self._app.debug:
            self._app["log"].write_message("commandline", command)
        # Save the cmd to the history list avoiding duplicates
        if command in self._history:
            self._history.remove(command)
        self._history.insert(0, command)
        # Reset history position
        self._pos = 0
        # Dereference aliases
        if command.lstrip(":") in self.commands.aliases:
            command = self.commands.aliases[command.lstrip(":")]
        # Run the correct command
        self._run(command)

    def _run(self, command):
        """Run correct function for command.

        Calls one of search, run_external_command, run_path and run_command.

        Args:
            command: The command to operate on.
        """
        # Search
        if command[0] == "/":
            self.search.run(command[1:])  # Strip /
        # Run something different depending on first char in command
        else:
            command = command.lstrip(":")
            # If there was no command just leave
            if not command:
                return
            elif command[0] == "!":
                self._run_external_command(command[1:])  # Strip the !
            elif command[0] == "~" or command[0] == "." or command[0] == "/":
                self._run_path(command)
            else:  # Default to internal command
                self._app["eventhandler"].num_clear()
                while command[0].isdigit():
                    self._app["eventhandler"].num_append(command[0])
                    command = command[1:]
                self.run_command(command)

    # Public because keyhandler also calls this
    def run_command(self, cmd, keyname=None):
        """Run the correct internal command.

        Args:
            cmd: Internal command to run.
            keyname: Key that called run_command. Passed when called from
                eventhandler.
        """
        # Get name and arguments
        name = cmd.split()[0]
        arguments = cmd.split()[1:]
        command_names = [cmd.name for cmd in self.commands]
        if name in command_names:
            command = self.commands[name]
            # Do not allow calling hidden commands from the command line
            if not keyname and command.is_hidden:
                message = "Called a hidden command from the command line"
                self._app["statusbar"].message(message, "error")
                return
            # Check if the command supports passing count
            if not command.supports_count:
                self._app["eventhandler"].num_clear()
            # Run command checking arguments
            try:
                command.run(arguments)
            except ArgumentAmountError as e:
                self._app["statusbar"].message(str(e), "error")
                # Check if the function supports passing count
        # If the command name is not in the dictionary throw an error
        else:
            message = "No command called " + name
            if keyname:
                message += " bound to " + keyname
            self._app["statusbar"].message(message, "error")

    def _run_external_command(self, cmd):
        """Run the entered command in the terminal.

        Args:
            cmd: The command to run.
        """
        # Expand filenames getting the correct filelist to operate on first
        fil = self._app.get_pos(True)
        if self._app["mark"].marked:  # Always use marked files if they exist
            filelist = list(self._app["mark"].marked)
        elif self._last_widget == "lib":
            filelist = list(self._app["library"].files)
        elif self._last_widget in ["thu", "im"]:
            filelist = list(self._app.get_paths())
        else:  # Empty filelist as a fallback
            filelist = []
        # Only do substitution if a filelist exists
        if filelist:
            cmd = expand_filenames(fil, filelist, cmd)
        # Possibility to "pipe to vimiv"
        if cmd[-1] == "|":
            cmd = cmd.rstrip("|")
            p = Popen(cmd, stdout=PIPE, stderr=PIPE, shell=True)
            from_pipe = True
        else:
            p = Popen(cmd, stderr=PIPE, shell=True)
            from_pipe = False
        self.running_processes.append(p)
        # Run the command in an extra thread, useful e.g. for gimp %
        cmd_thread = Thread(target=self._thread_for_external,
                            args=(cmd, p, from_pipe))
        cmd_thread.start()

    def _thread_for_external(self, cmd, p, from_pipe):
        """Start a new thread for external commands.

        Args:
            cmd: The command to run.
            directory: Directory which is affected by command.
        """
        # Get output and error and run the command
        out, err = p.communicate()
        if p.returncode:
            message = "Command exited with status " + str(p.returncode) + "\n"
            message += err.decode()
            GLib.idle_add(self._app["statusbar"].message, message, "error")
        else:
            # Run pipe if we have output
            if from_pipe:
                GLib.idle_add(self._run_pipe, out)
            # We do not know what might have changed concerning paths
            else:
                GLib.idle_add(self._app.emit, "paths-changed", self)
        self.running_processes.pop()

    def _run_pipe(self, pipe_input):
        """Run output of external command in a pipe.

        This checks for directories, files and vimiv commands.

        Args:
            pipe_input: Command that comes from pipe.
        """
        # List of commands without empty line
        pipe_input = pipe_input.decode("utf-8")
        pipe_input = pipe_input.split("\n")[:-1]
        startout = pipe_input[0] if pipe_input else ""
        # Leave if no pipe_input came
        if not startout or startout.isspace():
            self._app["statusbar"].message("No input from pipe", "info")
            return
        # Do different stuff depending on the first line of pipe_input
        if os.path.isdir(startout):
            self._app["library"].move_up(startout)
        elif os.path.isfile(startout):
            # Remember old file if no image was in filelist
            if self._app.get_paths():
                old_pos = self._app.get_index()
            else:
                old_pos = []
            # Populate filelist
            self._app.populate(pipe_input)
            if self._app.get_paths():  # Images were found
                self._app["main_window"].show()
                self._app["image"].load()
                # Resize library and focus image if necessary
                if self._app["library"].grid.is_visible():
                    self._app["library"].set_hexpand(False)
                    self._app["library"].focus(False)
            else:
                if old_pos:  # Back again if there is a remembered image
                    self._app.populate(old_pos)
                self._app["statusbar"].message("No image found", "info")
        else:  # Run every line as an internal command
            for cmd in pipe_input:
                self.run_command(cmd)

    def _run_path(self, path):
        """Run a path command, namely populate files or focus directory.

        Args:
            path: The path to run on.
        """
        # Expand home and use absolute path
        path = os.path.abspath(os.path.expanduser(path))
        if os.path.exists(path):
            if os.path.isdir(path):
                # Focus directory in the library
                self._app["library"].move_up(path)
                self._last_widget = "lib"
            else:
                # If it is an image open it
                self._app.populate([path])
                self._app["image"].load()
                #  Reload library in lib mode, do not open it in image mode
                pathdir = os.path.dirname(path)
                if self._last_widget == "lib":
                    self._app["library"].move_up(pathdir)
                    # Focus it in the treeview so it can be accessed via "l"
                    index = \
                        self._app["library"].files.index(os.path.basename(path))
                    self._app["library"].set_cursor(
                        Gtk.TreePath(index), None, False)
                    # Show the image
                    self._app["library"].set_hexpand(False)
                    self._app["main_window"].show()
                else:
                    self._app["library"].move_up(pathdir, True)
        else:
            self._app["statusbar"].message("Not a valid path", "error")

    def history_search(self, down):
        """Search through history with substring match.

        Updates the text of self with the matching history search.

        Args:
            down: If True, search downwards. Else search upwards.
        """
        # Shortly disconnect the change signal as it resets the history search
        self.disconnect_by_func(self._on_text_changed)
        # Only parts of the history that match the entered text
        if not self._matching_history:
            substring = self.get_text()
            matchstr = "^(" + substring + ")"
            self._matching_history = [substring]
            for cmd in self._history:
                if re.match(matchstr, cmd):
                    self._matching_history.append(cmd)
        # Move and set the text
        if down:
            self._pos -= 1
        else:
            self._pos += 1
        self._pos = self._pos % len(self._matching_history)
        self.set_text(self._matching_history[self._pos])
        self.set_position(-1)
        # Reconnect when done
        self.connect("changed", self._on_text_changed)

    def enter(self, text=""):
        """Prepend ':' to the cmd_line and open it.

        Args:
            text: Additionally add this to the command line.
        """
        self.set_text(":" + text)
        self._focus()

    def enter_search(self):
        """Prepend search character '/' to the cmd_line and open it."""
        focused_file = self._app.get_pos(True)
        self.set_text("/")
        self._focus()
        self.set_position(-1)
        if self._last_widget == "lib":
            paths = self._app["library"].files
        else:
            paths = [os.path.basename(path) for path in self._app.get_paths()]
        self.search.init(paths, focused_file, self._last_widget)

    def _focus(self):
        """Open and focus the command line."""
        # Show/hide the relevant stuff
        self.show()
        self._app["completions"].show()
        # Remember what widget was focused before
        self._last_widget = self._app.get_focused_widget()
        self.grab_focus()
        self.set_position(-1)
        # Update info for command mode
        self._app["statusbar"].update_info()

    def _on_text_changed(self, entry):
        """Close the entry if the colon/slash is deleted and reset search.

        Args:
            entry: The Gtk.Entry to check.
        """
        self._matching_history = []
        self._pos = 0
        text = entry.get_text()
        if not text or text[0] not in ":/":
            self.leave()

    def leave(self, reset_search=False):
        """Apply actions to close the commandline."""
        self._app["completions"].scroll_to_point(0, 0)
        self.hide()
        self._app["completions"].hide()
        self._app["completions"].reset()
        # Refocus the remembered widget
        if self._last_widget == "lib":
            self._app["library"].focus(True)
        elif self._last_widget == "man":
            self._app["manipulate"].sliders["bri"].grab_focus()
        elif self._last_widget == "thu":
            self._app["thumbnail"].grab_focus()
        else:
            self._app["main_window"].grab_focus()
        if reset_search:
            self.search.reset()
        self._last_index = 0
        self._app["statusbar"].update_info()

    def search_move(self, forward=True):
        """Move to the next or previous search.

        Args:
            forward: If true, move forwards. Else move backwards.
        """
        repeat = self._app["eventhandler"].num_receive()
        focused_file = os.path.basename(self._app.get_pos(True))
        try:
            self.search.next_result(forward, repeat, focused_file)
        except NoSearchResultsError as e:
            self._app["statusbar"].message(str(e), "error")

    def get_history(self):
        return self._history

    def write_history(self):
        """Write commandline history to file."""
        histfile = os.path.join(GLib.get_user_data_dir(), "vimiv", "history")
        with open(histfile, "w") as f:
            for cmd in self._history:
                f.write(cmd + "\n")

    def init_commands(self):
        """Initiate all commands as soon as the complete app is initialized."""
        self.commands = Commands(self._app)


class Search(GObject.Object):
    """Handle search for vimiv.

    Filelist gets set upon commandline entry. When search through filelist was
    completed, the "search-completed" signal gets emitted.

    Attributes:
        results: List of files in search results.

        _filelist: List of files to operate search on.
        _last_file: Filename that was focused before search.
        _last_widget: Widget that was focused before search.

    Signals:
        search-completed: Emitted when a search was completed so the widgets can
            react accordingly (focus new path, show search results, ...).
        no-search-results: Emitted when a search was result-less to show a
            warning.
    """

    def __init__(self, commandline):
        """Create necessary objects and settings.

        Args:
            commandline: Commandline to interact with.
        """
        super(Search, self).__init__()

        self._filelist = []
        self._last_file = ""
        self.results = []
        self._last_widget = ""

        if settings["incsearch"].get_value():
            commandline.connect("changed", self._do_incsearch)

    def _do_incsearch(self, entry):
        """Run a search for every entered character.

        Args:
            entry: The Gtk.Entry to check.
        """
        text = entry.get_text()
        if text[1:] and text[0] == "/" and self._last_widget != "im":
            self.run(text[1:])

    def init(self, filelist, focused_file, last_focused):
        """Initialize values needed for search.

        Called when entering the command line.
        """
        self._filelist = filelist
        self._last_file = os.path.basename(focused_file)
        self._last_widget = last_focused
        self.reset()

    def run(self, searchstr):
        """Start a search through filelist and emit search-completed signal."""
        case_sensitive = settings["search_case_sensitive"].get_value()
        self.results = \
            [fil for fil in self._filelist
             if searchstr in fil
             or not case_sensitive and searchstr.lower() in fil.lower()]
        if self.results:
            self.next_result(forward=True)
        else:
            self.emit("no-search-results", searchstr)

    def next_result(self, forward=True, repeat=1, focused_file=""):
        """Move to next search result by emitting search-completed.

        Args:
            forward: If True, search forward in list. Else backwards.
            repeat: Steps coming from eventhandler.
            focused_file: File that is focused before calling this. Required as
                this may be called with keybindings.
        """
        if focused_file:
            self._last_file = focused_file
        count = 0
        # Order filelist according to last position
        filelist = list(self._filelist)
        if not forward:
            filelist.reverse()
        if self._last_file not in filelist:
            raise NoSearchResultsError("No search results to navigate")
        index = filelist.index(self._last_file) + 1

        filelist = filelist[index:] + filelist[:index]
        # Find next match
        for i, f in enumerate(filelist):
            if f in self.results:
                count += 1
                if repeat == count:
                    new_pos = (index + i) if forward \
                        else len(self._filelist) - 1 - (index + i)
                    new_pos %= len(self._filelist)
                    self.emit("search-completed", new_pos, self._last_widget)
                    break
        return 0

    def reset(self):
        """Reset search and trigger reload of widgets."""
        last_pos = self._filelist.index(self._last_file) \
            if self._last_file in self._filelist else None
        self.results = []
        self.emit("search-completed", last_pos, self._last_widget)


# Initiate signals for the Search class
GObject.signal_new("search-completed", Search, GObject.SIGNAL_RUN_LAST,
                   None, (GObject.TYPE_PYOBJECT, GObject.TYPE_STRING))
GObject.signal_new("no-search-results", Search, GObject.SIGNAL_RUN_LAST,
                   None, (GObject.TYPE_PYOBJECT,))



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Window class for vimiv."""

from gi.repository import Gdk, Gtk
from vimiv.settings import settings


class Window(Gtk.ApplicationWindow):
    """"Vimiv Window containing the general structure.

    Additionally handles events like fullscreen and resize.

    Attributes:
        winsize: The windowsize as tuple.

        _app: The main vimiv application to interact with.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        super(Window, self).__init__()

        self._app = app

        self.connect("destroy", self._app.quit_wrapper)
        self.connect("button_press_event", self._app["eventhandler"].on_click,
                     "IMAGE")
        self.connect("touch-event", self._app["eventhandler"].on_touch)
        self.add_events(Gdk.EventMask.KEY_PRESS_MASK |
                        Gdk.EventMask.POINTER_MOTION_MASK)

        # Set initial window size
        self.winsize = settings["geometry"].get_value()
        self.resize(self.winsize[0], self.winsize[1])

        if settings["start_fullscreen"].get_value():
            self.toggle_fullscreen()

        # Auto resize
        self.connect("check-resize", self._auto_resize)
        # Focus changes with mouse
        for widget in [self._app["library"],
                       self._app["thumbnail"],
                       self._app["manipulate"].sliders["bri"],
                       self._app["manipulate"].sliders["con"],
                       self._app["manipulate"].sliders["sat"],
                       self._app["image"]]:
            widget.connect("button-release-event", self._on_click)

    def toggle_fullscreen(self):
        """Toggle fullscreen."""
        if self.get_window().get_state() & Gdk.WindowState.FULLSCREEN:
            self.unfullscreen()
        else:
            self.fullscreen()

    def zoom(self, zoom_in=True, step=1):
        """Zoom image or thumbnails.

        Args:
            zoom_in: If True, zoom in.
        """
        if self._app["thumbnail"].toggled:
            self._app["thumbnail"].zoom(zoom_in)
        else:
            self._app["image"].zoom_delta(zoom_in=zoom_in, step=step)

    def _auto_resize(self, window):
        """Automatically resize widgets when window is resized.

        Args:
            window: The window which emitted the resize event.
        """
        if self.get_size() != self.winsize:
            self.winsize = self.get_size()
            self._app.emit("widget-layout-changed", self)

    def _on_click(self, widget, event_button):
        """Update statusbar with the currently focused widget after mouse click.

        Args:
            widget: The widget that emitted the signal.
            event_button: Mouse button that was pressed.
        """
        self._app["statusbar"].update_info()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Slideshow for vimiv."""

from gi.repository import GLib, GObject
from vimiv.settings import settings


class Slideshow(GObject.Object):
    """Handle everything related to slideshow for vimiv.

    Attributes:
        running: If True the slideshow is running.

        _app: The main application class to interact with.
        _start_index: Index of the image when slideshow was started. Saved to
            display a message when slideshow is back at the beginning.
        _timer_id: ID of the currently running GLib.Timeout.
    """

    def __init__(self, app):
        """Create the necessary objects and settings.

        Args:
            app: The main vimiv application to interact with.
        """
        super(Slideshow, self).__init__()
        self._app = app

        self._start_index = 0
        self.running = False
        self._timer_id = GLib.Timeout
        settings.connect("changed", self._on_settings_changed)

    def toggle(self):
        """Toggle the slideshow or update the delay."""
        if not self._app.get_paths():
            message = "No valid paths, starting slideshow failed"
            self._app["statusbar"].message(message, "error")
            return
        if self._app["thumbnail"].toggled:
            message = "Slideshow makes no sense in thumbnail mode"
            self._app["statusbar"].message(message, "warning")
            return
        # Delay changed via vimiv["eventhandler"].num_str?
        number = self._app["eventhandler"].get_num_str()
        if number:
            settings.override("slideshow_delay", number)
            self._app["eventhandler"].num_clear()
        # If the delay wasn't changed in any way just toggle the slideshow
        else:
            self.running = not self.running
            if self.running:
                delay = 1000 * settings["slideshow_delay"].get_value()
                self._start_index = self._app.get_index()
                self._timer_id = GLib.timeout_add(delay, self._next)
            else:
                self._app["statusbar"].lock = False
                GLib.source_remove(self._timer_id)
        self._app["statusbar"].update_info()

    def _next(self):
        """Command to run in the GLib.timeout moving to the next image."""
        self._app["image"].move_index()
        # Info if slideshow returns to beginning
        if self._app.get_index() == self._start_index:
            message = "Back at beginning of slideshow"
            self._app["statusbar"].lock = True
            self._app["statusbar"].message(message, "info")
        else:
            self._app["statusbar"].lock = False
        return True  # So we continue running

    def get_formatted_delay(self):
        """Return the delay formatted neatly for the statusbar."""
        delay = settings["slideshow_delay"].get_value()
        return "[slideshow - {0:.1f}s]".format(delay)

    def _on_settings_changed(self, new_settings, setting):
        if setting == "slideshow_delay":
            delay = settings["slideshow_delay"].get_value()
            # Set a minimum
            if delay < 0.5:
                settings.override("slideshow_delay", "0.5")
                self._app["statusbar"].message(
                    "Delays shorter than 0.5 s are not allowed", "warning")
                return
            # If slideshow was running reload it
            if self.running:
                GLib.source_remove(self._timer_id)
                self._timer_id = GLib.timeout_add(1000 * delay, self._next)
                self._app["statusbar"].update_info()



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Handles the keyboard for vimiv."""

from gi.repository import Gdk, GLib, GObject
from vimiv.config_parser import parse_keys
from vimiv.helpers import get_float, get_int


class EventHandler(GObject.Object):
    """Handle keyboard/mouse/touch events for vimiv.

    Attributes:
        num_str: String containing repetition number for commands.

        _app: Main vimiv application to interact with.
        _keys: Keybindings from configfiles.
        _timer_id: ID of the current timer running to clear num_str.
        _timer_id_touch: ID of the current timer running to reconnect clicks
            after a touch by touchscreen.
    """

    def __init__(self, app):
        """Initialize defaults."""
        super(EventHandler, self).__init__()
        self._num_str = ""
        self._app = app
        self._timer_id = 0
        self._timer_id_touch = 0
        self._keys = parse_keys(running_tests=self._app.running_tests)

    def on_key_press(self, widget, event, widget_name):
        """Handle key press event."""
        if event.type == Gdk.EventType.KEY_PRESS:
            keyval = event.keyval
            keyname = Gdk.keyval_name(keyval)
        keyname = self._check_modifiers(event, keyname)
        return self._run(keyname, widget_name)

    def on_click(self, widget, event, widget_name):
        """Handle click event."""
        # Do not handle double clicks
        if not event.type == Gdk.EventType.BUTTON_PRESS:
            return
        button_name = "Button" + str(event.button)
        button_name = self._check_modifiers(event, button_name)
        return self._run(button_name, widget_name)

    def _run(self, keyname, widget_name):
        """Run the correct function per keypress.

        Args:
            keyname: Name of the key that was pressed/clicked/...
            widget_name: Name of widget to operate on: image, library...
        """
        # Numbers for the num_str
        if widget_name != "COMMAND" and keyname.isdigit():
            self.num_append(keyname)
            return True
        # Get the relevant keybindings for the window from the various
        # sections in the keys.conf file
        keys = self._keys[widget_name]
        # Get the command to which the pressed key is bound and run it
        if keyname in keys:
            keybinding = keys[keyname]
            # Write keybinding and key to log in debug mode
            if self._app.debug:
                self._app["log"].write_message("key",
                                               keyname + ": " + keybinding)
            self._app["commandline"].run_command(keybinding, keyname)
            return True  # Deactivates default bindings
        # Activate default keybindings
        else:
            return False

    def on_touch(self, widget, event):
        """Clear mouse connection when touching screen.

        This stops calling the ButtonX bindings when using the touch screen.
        Reasoning: We do not want to e.g. move to the next image when trying to
            zoom in.
        """
        try:
            self._app["window"].disconnect_by_func(self.on_click)
        # Was already disconnected
        except TypeError:
            pass
        if self._timer_id_touch:
            GLib.source_remove(self._timer_id_touch)
        self._timer_id_touch = GLib.timeout_add(5, self._reconnect_click)
        return True

    def _reconnect_click(self):
        """Reconnect the click signal after a touch event."""
        self._app["window"].connect("button_press_event",
                                    self.on_click, "IMAGE")
        self._timer_id_touch = 0

    def _check_modifiers(self, event, keyname):
        """Update keyname according to modifiers in event."""
        shiftkeys = ["space", "Return", "Tab", "Escape", "BackSpace",
                     "Up", "Down", "Left", "Right"]
        # Check for Control (^), Mod1 (Alt) or Shift
        if event.get_state() & Gdk.ModifierType.CONTROL_MASK:
            keyname = "^" + keyname
        if event.get_state() & Gdk.ModifierType.MOD1_MASK:
            keyname = "Alt+" + keyname
        # Shift+ for all letters and for keys that don't support it
        if (event.get_state() & Gdk.ModifierType.SHIFT_MASK and
                (len(keyname) < 2 or keyname in shiftkeys
                 or keyname.startswith("Button"))):
            keyname = "Shift+" + keyname.lower()
        if keyname == "ISO_Left_Tab":  # Tab is named really weird under shift
            keyname = "Shift+Tab"
        return keyname

    def num_append(self, num, remove_by_timeout=True):
        """Add a new char to num_str.

        Args:
            num: The number to append to the string.
            remove_by_timeout: If True, add a timeout to clear the num_str.
        """
        # Remove old timers if we have new numbers
        if self._timer_id:
            GLib.source_remove(self._timer_id)
        self._timer_id = GLib.timeout_add_seconds(1, self.num_clear)
        self._num_str += num
        self._convert_trailing_zeros()
        # Write number to log file in debug mode
        if self._app.debug:
            self._app["log"].write_message("number", num + "->" + self._num_str)
        self._app["statusbar"].update_info()

    def num_clear(self):
        """Clear num_str."""
        # Remove any timers as we are clearing now anyway
        if self._timer_id:
            GLib.source_remove(self._timer_id)
        # Write number cleared to log file in debug mode
        if self._app.debug and self._num_str:
            self._app["log"].write_message("number", "cleared")
        self._timer_id = 0
        # Reset
        self._num_str = ""
        self._app["statusbar"].update_info()

    def num_receive(self, number=1, to_float=False):
        """Receive self._num_str and clear it.

        Args:
            number: Number to return if self._num_str is empty.
            to_float: If True, convert num_str to float. Else to int.
        Return:
            The received number or default.
        """
        if self._num_str:
            number = get_float(self._num_str) \
                if to_float else get_int(self._num_str)
            self.num_clear()
        return number

    def get_num_str(self):
        return self._num_str

    def set_num_str(self, number):
        self.num_clear()
        self.num_append(str(number))

    def _convert_trailing_zeros(self):
        """If prefixed with zero add a decimal point to self._num_str."""
        if self._num_str.startswith("0") and not self._num_str.startswith("0."):
            self._num_str = self._num_str.replace("0", "0.")



# vim: ft=python fileencoding=utf-8 sw=4 et sts=4
"""Statusbar for vimiv."""

import os

from gi.repository import Gdk, GLib, Gtk
from vimiv.settings import settings


class Statusbar(Gtk.Grid):
    """Create the statusbar and handle all events for it.

    Attributes:
        lock: If True, do not update any information.
        separator: Gtk.Separator used as background of the statusbar. Makes sure
            the other widgets do not interfere with the overlaying bar.

        _app: The main vimiv application to interact with.
        _center_label: Gtk.Label containing mark status and slideshow info.
        _errors: True if there are errors.
        _left_label: Gtk.Label containing position, name and zoom.
        _right_label: Gtk.Label containing mode and prefixed numbers.
        _timer_id: ID of the currently running GLib.Timeout.
        _was_hidden: If True the statusbar was hidden before an error message.
    """

    def __init__(self, app):
        super(Statusbar, self).__init__()
        self._app = app

        # Default values
        self.lock = False
        self._errors = True
        self._timer_id = 0
        self._was_hidden = False

        # Statusbar on the bottom
        self.set_name("StatusBar")  # Name for css
        # Two labels for two sides of statusbar and one in the middle for
        # additional info
        self._left_label = Gtk.Label()
        self._left_label.set_justify(Gtk.Justification.LEFT)
        self._right_label = Gtk.Label()
        self._right_label.set_justify(Gtk.Justification.RIGHT)
        self._center_label = Gtk.Label()
        self._center_label.set_justify(Gtk.Justification.CENTER)
        self._center_label.set_hexpand(True)
        # Add them all
        self.attach(self._left_label, 0, 0, 1, 1)
        self.attach(self._center_label, 1, 0, 1, 1)
        self.attach(self._right_label, 2, 0, 1, 1)
        # Padding and separator
        padding = settings["commandline_padding"].get_value()
        self.set_margin_start(padding)
        self.set_margin_end(padding)
        self.set_margin_top(padding)
        self.set_margin_bottom(padding)
        self.separator = Gtk.Separator()

        # Connect signals
        self._app["commandline"].search.connect("no-search-results",
                                                self._on_no_search_results)
        settings.connect("changed", self._on_settings_changed)

    def message(self, message, style="error", timeout=5):
        """Push a message to the statusbar.

        Args:
            message: Message to push.
            style: One of error, warning and info.
            timeout: Time until the message shall be removed by update info.
        """
        if self._app.debug:
            self._app["log"].write_message(style, message)
        self._errors = True
        styles = {"error": "#CC0000", "warning": "#FA9E21", "info": "#6699FF"}
        message = "<b><span foreground='" + styles[style] + "'>" + \
            style.upper() + ": </span></b>" + message
        self._timer_id = GLib.timeout_add(timeout * 1000, self.update_info)
        # Show if is was hidden
        if not settings["display_bar"].get_value():
            settings.override("display_bar", "true")
            self._was_hidden = True
        self._left_label.set_markup(message)
        # CSS to style bar according to message
        css_provider = Gtk.CssProvider()
        css_str = "#OverLay { border-top: solid 2px " + styles[style] + ";}"
        css_provider.load_from_data(css_str.encode())
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def update_info(self):
        """Update the statusbar and the window title."""
        # Return if it is locked
        if self.lock:
            return
        # Hide again if it was shown due to an error message
        if self._was_hidden:
            self._was_hidden = False
            settings.override("display_bar", "false")
        # Strip error messages if any
        if self._errors:
            self._error_false()
        # Get mode first
        mode = self._get_mode()
        # Update all relevant widgets
        self._set_left_status(mode)
        self._set_center_status(mode)
        self._set_right_status(mode)
        self._set_window_title()

    def _error_false(self):
        """Strip one error and update the statusbar if no more errors remain."""
        # Remove any timers that remove the error message
        if self._timer_id:
            GLib.source_remove(self._timer_id)
            self._timer_id = 0
        # Strip one error
        self._errors = False
        # Reset css from error_message
        css_provider = Gtk.CssProvider()
        css_str = "#OverLay { border-top: solid 0px; }"
        css_provider.load_from_data(css_str.encode())
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(), css_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

    def _set_left_status(self, mode):
        """Set the left side of the statusbar depending on mode."""
        # Directory if library is focused
        if "LIBRARY" in mode:
            cur_dir = os.getcwd()
            if settings["tilde_in_statusbar"].get_value():
                cur_dir = cur_dir.replace(os.getenv("HOME"), "~")
            self._left_label.set_text(cur_dir)
        # Position, name and thumbnail size in thumb mode
        elif "THUMBNAIL" in mode:
            pos = self._app.get_pos()
            name = os.path.basename(self._app.get_pos(True))
            message = "{0}/{1}  {2}  {3}".format(
                pos + 1, len(self._app.get_paths()), name,
                self._app["thumbnail"].get_zoom_level())

            self._left_label.set_text(message)
        # In commandline
        elif "COMMAND" in mode:
            # TODO useful information in the commandline
            self._left_label.set_text("")
        elif self._app.get_paths():
            name = os.path.basename(self._app.get_path())
            message = "{0}/{1}  {2}  [{3:.0f}%]".format(
                self._app.get_index() + 1, len(self._app.get_paths()), name,
                self._app["image"].zoom_percent * 100)
            self._left_label.set_text(message)
        else:
            self._left_label.set_text("No open images")

    def _set_center_status(self, mode):
        """Set the centre of the statusbar depending on mode."""
        mark = "[*]" \
            if ("IMAGE" in mode or "MANIPULATE" in mode) \
            and self._app.get_paths() \
            and self._app.get_path() in self._app["mark"].marked \
            else ""
        slideshow = self._app["slideshow"].get_formatted_delay() \
            if self._app["slideshow"].running else ""
        message = "{0}  {1}".format(mark, slideshow)
        self._center_label.set_text(message)

    def _set_right_status(self, mode):
        """Set the right side of the statusbar to mode and num_str."""
        message = "{0:15}  {1:4}".format(
            mode, self._app["eventhandler"].get_num_str())
        self._right_label.set_markup(message)

    def _set_window_title(self):
        """Set window title depending on whether there are valid paths."""
        if self._app.get_paths():
            name = os.path.basename(self._app.get_path())
            self._app["window"].set_title("vimiv - " + name)
        else:
            self._app["window"].set_title("vimiv")

    def _get_mode(self):
        """Return which widget is currently focused."""
        if self._app["commandline"].is_focus():
            return "<b>-- COMMAND --</b>"
        elif self._app["library"].is_focus():
            return "<b>-- LIBRARY --</b>"
        elif self._app["manipulate"].is_visible():
            return "<b>-- MANIPULATE --</b>"
        elif self._app["thumbnail"].toggled:
            return "<b>-- THUMBNAIL --</b>"
        return "<b>-- IMAGE --</b>"

    def clear_status(self):
        """Clear num_str, search and error messages from the statusbar."""
        self._app["commandline"].search.reset()
        self._app["eventhandler"].num_clear()

    def set_separator_height(self):
        """Set height of the separator used as background of the statusbar."""
        self.separator.set_size_request(1, self.get_bar_height())

    def get_bar_height(self):
        """Return height of the statusbar + padding."""
        bar_height = self.get_allocated_height()
        padding = settings["commandline_padding"].get_value()
        return bar_height + 2 * padding

    def get_message(self):
        return self._left_label.get_text()

    def _on_no_search_results(self, search, searchstr):
        self.message('No file matching "%s"' % (searchstr), "info")

    def _on_settings_changed(self, new_settings, setting):
        if setting == "display_bar":
            if settings["display_bar"].get_value():
                self.show()
                self.separator.show()
            elif not self._app["commandline"].is_visible():
                self.hide()
                self.separator.hide()
            self._app.emit("widget-layout-changed", self)
